/* packet-ocfs2.c
 *
 * Routines for OCFS2's networking protool disassembly
 *
 * $Id:$ packet-ocfs2.c
 *
 * Copyright 2006, Oracle Inc.
 * Authors: Kurt Hackel, Zach Brown, Sunil Mushran
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * XXX:
 * 	- find_header could require min header len
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>

#include <glib.h>
#include <epan/packet.h>
#include <epan/ipproto.h>
#include <epan/addr_resolv.h>
#include <epan/emem.h>
#include "packet-tcp.h"
#include <epan/prefs.h>
#include <epan/strutil.h>
#include <epan/emem.h>

#if (BITS_PER_LONG == 32) || defined(CONFIG_X86_64)
#define MLFi64 "lld"
#define MLFu64 "llu"
#define MLFx64 "llx"
#else
#define MLFi64 "ld"
#define MLFu64 "lu"
#define MLFx64 "lx"
#endif

#define FIELD_OFFSET_AND_SIZE(S,F)   offsetof(S,F), sizeof(((S*)0L)->F)

static gint ett_ocfs2 = -1;
static gint ett_dlm_master_request = -1;
static gint ett_dlm_assert_master = -1;
static gint ett_dlm_create_lock = -1;
static gint ett_dlm_convert_lock = -1;
static gint ett_dlm_proxy_ast = -1;
static gint ett_dlm_unlock_lock = -1;
static gint ett_dlm_migrate_request = -1;
static gint ett_dlm_migrate_lockres = -1;
static gint ett_dlm_query_join = -1;
static gint ett_dlm_assert_joined = -1;
static gint ett_dlm_cancel_join = -1;
static gint ett_dlm_exit_domain = -1;
static gint ett_dlm_master_requery = -1;
static gint ett_dlm_lock_request = -1;
static gint ett_dlm_recovery_data_done = -1;
static gint ett_dlm_begin_recovery = -1;
static gint ett_dlm_finalize_recovery = -1;

static int proto_ocfs2 = -1;
static int hf_msg_magic = -1;
static int hf_msg_data_len = -1;
static int hf_msg_msg_type = -1;
static int hf_msg_pad1 = -1;
static int hf_msg_sys_status = -1;
static int hf_msg_status = -1;
static int hf_msg_key = -1;
static int hf_msg_msg_num = -1;
static int hf_msg_payload = -1;

static int hf_dlm_node_idx = -1;
static int hf_dlm_flags = -1;
static int hf_dlm_am_flags = -1;
static int hf_dlm_fr_flags = -1;
static int hf_dlm_namelen = -1;
static int hf_dlm_name = -1;
static int hf_dlm_cookie = -1;
static int hf_dlm_requested_type = -1;
static int hf_dlm_lvb = -1;
static int hf_dlm_ast_type = -1;
static int hf_dlm_blocked_type = -1;
static int hf_dlm_dead_node = -1;
static int hf_dlm_domain_name_len = -1;
static int hf_dlm_domain_name = -1;

/* ocfs2 lock name expansion */
static int hf_dlm_ocfs2_blkno = -1;
static int hf_dlm_ocfs2_lock_type = -1;
static int hf_dlm_ocfs2_generation = -1;

/* ocfs2 lvb expansion */
static int hf_dlm_ocfs2_clusters = -1;
static int hf_dlm_ocfs2_size = -1;
static int hf_dlm_ocfs2_uid = -1;
static int hf_dlm_ocfs2_gid = -1;
static int hf_dlm_ocfs2_mode = -1;
static int hf_dlm_ocfs2_nlink = -1;
static int hf_dlm_ocfs2_atime = -1;
static int hf_dlm_ocfs2_ctime = -1;
static int hf_dlm_ocfs2_mtime = -1;


#define TCP_PORT_OCFS2	7777
#define O2NM_MAX_NAME_LEN        64
#define	O2NET_MSG_HDR_LEN		(2+2+2+2+4+4+4+4)


struct ocfs2_msg
{
        guint16 magic;
        guint16 data_len;
        guint16 msg_type;
        guint16 pad1;
        guint32 sys_status;
        guint32 status;
        guint32 key;
        guint32 msg_num;
        guint8  buf[0];
};
#define	O2NET_MSG_HDR_OFF_MAGIC		offsetof(struct ocfs2_msg,magic)
#define	O2NET_MSG_HDR_OFF_DATA_LEN	offsetof(struct ocfs2_msg,data_len)
#define	O2NET_MSG_HDR_OFF_MSG_TYPE	offsetof(struct ocfs2_msg,msg_type)
#define	O2NET_MSG_HDR_OFF_PAD1		offsetof(struct ocfs2_msg,pad1)
#define	O2NET_MSG_HDR_OFF_SYS_STATUS	offsetof(struct ocfs2_msg,sys_status)
#define	O2NET_MSG_HDR_OFF_STATUS	offsetof(struct ocfs2_msg,status)
#define	O2NET_MSG_HDR_OFF_KEY		offsetof(struct ocfs2_msg,key)
#define	O2NET_MSG_HDR_OFF_MSG_NUM	offsetof(struct ocfs2_msg,msg_num)
#define	O2NET_MSG_HDR_OFF_PAYLOAD	offsetof(struct ocfs2_msg,buf)


/*
 * generic o2net constants
 */

#define O2NET_MSG_MAGIC			0xfa55
#define O2NET_MSG_STATUS_MAGIC		0xfa56
#define O2NET_MSG_KEEP_REQ_MAGIC  	0xfa57
#define O2NET_MSG_KEEP_RESP_MAGIC 	0xfa58
static const value_string o2net_magic[] = {
	{ O2NET_MSG_MAGIC, "Request" },
	{ O2NET_MSG_STATUS_MAGIC, "Response" },
	{ O2NET_MSG_KEEP_REQ_MAGIC, "Keepalive Request" },
	{ O2NET_MSG_KEEP_RESP_MAGIC, "Keepalive Response" },
	{ 0x0000,  NULL  }
};


enum ocfs2_lock_type {
	OCFS2_LOCK_TYPE_INVALID = -1,
	OCFS2_LOCK_TYPE_META = 0,
	OCFS2_LOCK_TYPE_DATA,
	OCFS2_LOCK_TYPE_SUPER,
	OCFS2_LOCK_TYPE_RENAME,
};

static const value_string ocfs2_lock_types[] = {
	{ OCFS2_LOCK_TYPE_META, "Metadata" },
	{ OCFS2_LOCK_TYPE_DATA, "Data" },
	{ OCFS2_LOCK_TYPE_SUPER, "Superblock" },
	{ OCFS2_LOCK_TYPE_RENAME, "Rename" },
	{ 0x0000,  NULL  }
};

static inline enum ocfs2_lock_type ocfs2_lock_type_from_lockname(char *lockname)
{
	enum ocfs2_lock_type ret;
	char c = *lockname;

	switch (c) {
		case 'M':
			ret = OCFS2_LOCK_TYPE_META;
			break;
		case 'D':
			ret = OCFS2_LOCK_TYPE_DATA;
			break;
		case 'S':
			ret = OCFS2_LOCK_TYPE_SUPER;
			break;
		case 'R':
			ret = OCFS2_LOCK_TYPE_RENAME;
			break;
		default:
			ret = OCFS2_LOCK_TYPE_INVALID;
			break;
	}
	return ret;
}

static inline const char * ocfs2_lock_type_to_str(enum ocfs2_lock_type type)
{
	switch (type) {
		case OCFS2_LOCK_TYPE_META:
			return "Metadata";
		case OCFS2_LOCK_TYPE_DATA:
			return "Data";
		case OCFS2_LOCK_TYPE_SUPER:
			return "Superblock";
		case OCFS2_LOCK_TYPE_RENAME:
			return "Rename";
		default:
			return "Unknown";
	}
	return "Unknown";
}




#define OCFS2_SEC_SHIFT  (64 - 34)
#define OCFS2_NSEC_MASK  ((1ULL << OCFS2_SEC_SHIFT) - 1)

static void ocfs2_lvb_time_get_localtime(guint64 packed_time, struct tm *lt)
{
	const time_t t = packed_time >> OCFS2_SEC_SHIFT;
	localtime_r(&t, lt);
}

static guint32 ocfs2_lvb_time_get_nsec(guint64 packed_time)
{
	return packed_time & OCFS2_NSEC_MASK;
}


const char * decode_enumerated_bitfield_full(guint32 val, guint32 mask, int width, const value_string *tab, const char *fmt)
{
	static char buf[1025];
	char *p;
	int shift = 0;
	int len;

	p = decode_bitfield_value(buf, val, mask, width);

	while (1) {
		if (shift >= width)
			break;
		while ((mask & (1<<shift)) == 0)
			shift++;
		if (shift >= width)
			break;
		len = g_snprintf(p, 1024-(p-buf), fmt, val_to_str(val & (1<<shift), tab, ""));
		p += len;
		if (len)
			p += g_snprintf(p, 1024-(p-buf), " ");
		shift++;
	}
	return buf;
}


/*
 * DLM constants
 */

#define DLM_LVB_LEN  64
#define DLM_MOD_KEY (0x666c6172)

enum dlm_query_join_response {
	JOIN_DISALLOW = 0,
	JOIN_OK,
	JOIN_OK_NO_MAP,
};


/* DLM lock modes */
enum {
	LKM_IVMODE = -1,
	LKM_NLMODE = 0,
	LKM_CRMODE,
	LKM_CWMODE,
	LKM_PRMODE,
	LKM_PWMODE,
	LKM_EXMODE,
};
static const value_string dlm_lock_modes[] = {
	{ LKM_IVMODE, "IV: Invalid mode" },
	{ LKM_NLMODE, "NL: No lock" },
	{ LKM_CRMODE, "CR: Concurrent read (UNSUPPORTED)" },
	{ LKM_CWMODE, "CW: Concurrent write (UNSUPPORTED)" },
	{ LKM_PRMODE, "PR: Protected read" },
	{ LKM_PWMODE, "PW: Protected write (UNSUPPORTED)" },
	{ LKM_EXMODE, "EX: Exclusive" },
	{ 0x0000,  NULL  }
};


/* DLM message types */
enum {
	OCFS2_MESSAGE_TYPE_VOTE   = 1,
	OCFS2_MESSAGE_TYPE_RESPONSE = 2,
        DLM_MASTER_REQUEST_MSG    = 500,
        DLM_UNUSED_MSG1,
        DLM_ASSERT_MASTER_MSG,
        DLM_CREATE_LOCK_MSG,
        DLM_CONVERT_LOCK_MSG,
        DLM_PROXY_AST_MSG,
        DLM_UNLOCK_LOCK_MSG,
        DLM_UNUSED_MSG2,
        DLM_MIGRATE_REQUEST_MSG,
        DLM_MIG_LOCKRES_MSG,
        DLM_QUERY_JOIN_MSG,
        DLM_ASSERT_JOINED_MSG,
        DLM_CANCEL_JOIN_MSG,
        DLM_EXIT_DOMAIN_MSG,
        DLM_MASTER_REQUERY_MSG,
        DLM_LOCK_REQUEST_MSG,
        DLM_RECO_DATA_DONE_MSG,
        DLM_BEGIN_RECO_MSG,
        DLM_FINALIZE_RECO_MSG
};
static const value_string dlm_magic[] = {
	{ OCFS2_MESSAGE_TYPE_VOTE, "FS Vote" },
	{ OCFS2_MESSAGE_TYPE_RESPONSE, "FS Response" }, 
	{ DLM_MASTER_REQUEST_MSG, "Master Request" },
	{ DLM_UNUSED_MSG1,        "Unused 1" },
	{ DLM_ASSERT_MASTER_MSG,  "Assert Master" },
	{ DLM_CREATE_LOCK_MSG,    "Create Lock" },
	{ DLM_CONVERT_LOCK_MSG,   "Convert Lock" },
	{ DLM_PROXY_AST_MSG,      "Proxy AST" },
	{ DLM_UNLOCK_LOCK_MSG,    "Unlock Lock" },
	{ DLM_UNUSED_MSG2,        "Unused 2" },
	{ DLM_MIGRATE_REQUEST_MSG,"Migrate Request" },
	{ DLM_MIG_LOCKRES_MSG,    "Migrate Lockres" },
	{ DLM_QUERY_JOIN_MSG,     "Query Join" },
	{ DLM_ASSERT_JOINED_MSG,  "Assert Join" },
	{ DLM_CANCEL_JOIN_MSG,    "Cancel Join" },
	{ DLM_EXIT_DOMAIN_MSG,    "Exit Domain" },
	{ DLM_MASTER_REQUERY_MSG, "Master Requery" },
	{ DLM_LOCK_REQUEST_MSG,   "Lock Request" },
	{ DLM_RECO_DATA_DONE_MSG, "Recovery Data Done" },
	{ DLM_BEGIN_RECO_MSG,     "Begin Reovery" },
	{ DLM_FINALIZE_RECO_MSG,  "Finalize Recovery" },
	{ 0x0000, NULL }
};


/* DLM lock flags */
#define LKM_ORPHAN       0x00000010
#define LKM_PARENTABLE   0x00000020
#define LKM_BLOCK        0x00000040
#define LKM_LOCAL        0x00000080
#define LKM_VALBLK       0x00000100
#define LKM_NOQUEUE      0x00000200
#define LKM_CONVERT      0x00000400
#define LKM_NODLCKWT     0x00000800
#define LKM_UNLOCK       0x00001000
#define LKM_CANCEL       0x00002000
#define LKM_DEQALL       0x00004000
#define LKM_INVVALBLK    0x00008000
#define LKM_SYNCSTS      0x00010000
#define LKM_TIMEOUT      0x00020000
#define LKM_SNGLDLCK     0x00040000
#define LKM_FINDLOCAL    0x00080000
#define LKM_PROC_OWNED   0x00100000
#define LKM_XID          0x00200000
#define LKM_XID_CONFLICT 0x00400000
#define LKM_FORCE        0x00800000
#define LKM_REVVALBLK    0x01000000
#define LKM_UNUSED1      0x00000001
#define LKM_UNUSED2      0x00000002
#define LKM_UNUSED3      0x00000004
#define LKM_UNUSED4      0x00000008
#define LKM_UNUSED5      0x02000000
#define LKM_UNUSED6      0x04000000
#define LKM_UNUSED7      0x08000000
#define LKM_MIGRATION    0x10000000
#define LKM_PUT_LVB      0x20000000
#define LKM_GET_LVB      0x40000000
#define LKM_RECOVERY     0x80000000

static const value_string dlm_lock_flags[] = {
	{ LKM_ORPHAN, "orphan" },
	{ LKM_PARENTABLE, "parentable" },
	{ LKM_BLOCK, "block" },
	{ LKM_LOCAL, "local" },
	{ LKM_VALBLK, "valblk" },
	{ LKM_NOQUEUE, "noqueue" },
	{ LKM_CONVERT, "convert" },
	{ LKM_NODLCKWT, "nodlckwt" },
	{ LKM_UNLOCK, "unlock" },
	{ LKM_CANCEL, "cancel" },
	{ LKM_DEQALL, "deqall" },
	{ LKM_INVVALBLK, "invvalblk" },
	{ LKM_SYNCSTS, "syncsts" },
	{ LKM_TIMEOUT, "timeout" },
	{ LKM_SNGLDLCK, "sngldlck" },
	{ LKM_FINDLOCAL, "findlocal" },
	{ LKM_PROC_OWNED, "proc_owned" },
	{ LKM_XID, "xid" },
	{ LKM_XID_CONFLICT, "xid_conflict" },
	{ LKM_FORCE, "force" },
	{ LKM_REVVALBLK, "revvalblk" },
	{ LKM_UNUSED1, "unused1" },
	{ LKM_UNUSED2, "unused2" },
	{ LKM_UNUSED3, "unused3" },
	{ LKM_UNUSED4, "unused4" },
	{ LKM_UNUSED5, "unused5" },
	{ LKM_UNUSED6, "unused6" },
	{ LKM_UNUSED7, "unused7" },
	{ LKM_MIGRATION, "migration" },
	{ LKM_PUT_LVB, "put_lvb" },
	{ LKM_GET_LVB, "get_lvb" },
	{ LKM_RECOVERY, "recovery" },
	{ 0x0000,  NULL }
};

enum dlm_status {
	DLM_NORMAL = 0,
	DLM_GRANTED,
	DLM_DENIED,
	DLM_DENIED_NOLOCKS,
	DLM_WORKING,
	DLM_BLOCKED,
	DLM_BLOCKED_ORPHAN,
	DLM_DENIED_GRACE_PERIOD,
	DLM_SYSERR,
	DLM_NOSUPPORT,
	DLM_CANCELGRANT,
	DLM_IVLOCKID,
	DLM_SYNC,
	DLM_BADTYPE,
	DLM_BADRESOURCE,
	DLM_MAXHANDLES,
	DLM_NOCLINFO,
	DLM_NOLOCKMGR,
	DLM_NOPURGED,
	DLM_BADARGS,
	DLM_VOID,
	DLM_NOTQUEUED,
	DLM_IVBUFLEN,
	DLM_CVTUNGRANT,
	DLM_BADPARAM,
	DLM_VALNOTVALID,
	DLM_REJECTED,
	DLM_ABORT,
	DLM_CANCEL,
	DLM_IVRESHANDLE,
	DLM_DEADLOCK,
	DLM_DENIED_NOASTS,
	DLM_FORWARD,
	DLM_TIMEOUT,
	DLM_IVGROUPID,
	DLM_VERS_CONFLICT,
	DLM_BAD_DEVICE_PATH,
	DLM_NO_DEVICE_PERMISSION,
	DLM_NO_CONTROL_DEVICE,
	DLM_RECOVERING,
	DLM_MIGRATING,
	DLM_MAXSTATS,
};

static const value_string dlm_errnames[] = {
	{ DLM_NORMAL, "DLM_NORMAL" },
	{ DLM_GRANTED, "DLM_GRANTED" },
	{ DLM_DENIED, "DLM_DENIED" },
	{ DLM_DENIED_NOLOCKS, "DLM_DENIED_NOLOCKS" },
	{ DLM_WORKING, "DLM_WORKING" },
	{ DLM_BLOCKED, "DLM_BLOCKED" },
	{ DLM_BLOCKED_ORPHAN, "DLM_BLOCKED_ORPHAN" },
	{ DLM_DENIED_GRACE_PERIOD, "DLM_DENIED_GRACE_PERIOD" },
	{ DLM_SYSERR, "DLM_SYSERR" },
	{ DLM_NOSUPPORT, "DLM_NOSUPPORT" },
	{ DLM_CANCELGRANT, "DLM_CANCELGRANT" },
	{ DLM_IVLOCKID, "DLM_IVLOCKID" },
	{ DLM_SYNC, "DLM_SYNC" },
	{ DLM_BADTYPE, "DLM_BADTYPE" },
	{ DLM_BADRESOURCE, "DLM_BADRESOURCE" },
	{ DLM_MAXHANDLES, "DLM_MAXHANDLES" },
	{ DLM_NOCLINFO, "DLM_NOCLINFO" },
	{ DLM_NOLOCKMGR, "DLM_NOLOCKMGR" },
	{ DLM_NOPURGED, "DLM_NOPURGED" },
	{ DLM_BADARGS, "DLM_BADARGS" },
	{ DLM_VOID, "DLM_VOID" },
	{ DLM_NOTQUEUED, "DLM_NOTQUEUED" },
	{ DLM_IVBUFLEN, "DLM_IVBUFLEN" },
	{ DLM_CVTUNGRANT, "DLM_CVTUNGRANT" },
	{ DLM_BADPARAM, "DLM_BADPARAM" },
	{ DLM_VALNOTVALID, "DLM_VALNOTVALID" },
	{ DLM_REJECTED, "DLM_REJECTED" },
	{ DLM_ABORT, "DLM_ABORT" },
	{ DLM_CANCEL, "DLM_CANCEL" },
	{ DLM_IVRESHANDLE, "DLM_IVRESHANDLE" },
	{ DLM_DEADLOCK, "DLM_DEADLOCK" },
	{ DLM_DENIED_NOASTS, "DLM_DENIED_NOASTS" },
	{ DLM_FORWARD, "DLM_FORWARD" },
	{ DLM_TIMEOUT, "DLM_TIMEOUT" },
	{ DLM_IVGROUPID, "DLM_IVGROUPID" },
	{ DLM_VERS_CONFLICT, "DLM_VERS_CONFLICT" },
	{ DLM_BAD_DEVICE_PATH, "DLM_BAD_DEVICE_PATH" },
	{ DLM_NO_DEVICE_PERMISSION, "DLM_NO_DEVICE_PERMISSION" },
	{ DLM_NO_CONTROL_DEVICE , "DLM_NO_CONTROL_DEVICE " },
	{ DLM_RECOVERING, "DLM_RECOVERING" },
	{ DLM_MIGRATING, "DLM_MIGRATING" },
	{ DLM_MAXSTATS, "DLM_MAXSTATS" },
	{ 0x0000,  NULL }
};

static const value_string dlm_errmsgs[] = {
	{ DLM_NORMAL, "request in progress" },
	{ DLM_GRANTED, "request granted" },
	{ DLM_DENIED, "request denied" },
	{ DLM_DENIED_NOLOCKS, "request denied, out of system resources" },
	{ DLM_WORKING, "async request in progress" },
	{ DLM_BLOCKED, "lock request blocked" },
	{ DLM_BLOCKED_ORPHAN, "lock request blocked by a orphan lock" },
	{ DLM_DENIED_GRACE_PERIOD, "topological change in progress" },
	{ DLM_SYSERR, "system error" },
	{ DLM_NOSUPPORT, "unsupported" },
	{ DLM_CANCELGRANT, "can't cancel convert: already granted" },
	{ DLM_IVLOCKID, "bad lockid" },
	{ DLM_SYNC, "synchronous request granted" },
	{ DLM_BADTYPE, "bad resource type" },
	{ DLM_BADRESOURCE, "bad resource handle" },
	{ DLM_MAXHANDLES, "no more resource handles" },
	{ DLM_NOCLINFO, "can't contact cluster manager" },
	{ DLM_NOLOCKMGR, "can't contact lock manager" },
	{ DLM_NOPURGED, "can't contact purge daemon" },
	{ DLM_BADARGS, "bad api args" },
	{ DLM_VOID, "no status" },
	{ DLM_NOTQUEUED, "NOQUEUE was specified and request failed" },
	{ DLM_IVBUFLEN, "invalid resource name length" },
	{ DLM_CVTUNGRANT, "attempted to convert ungranted lock" },
	{ DLM_BADPARAM, "invalid lock mode specified" },
	{ DLM_VALNOTVALID, "value block has been invalidated" },
	{ DLM_REJECTED, "request rejected, unrecognized client" },
	{ DLM_ABORT, "blocked lock request cancelled" },
	{ DLM_CANCEL, "conversion request cancelled" },
	{ DLM_IVRESHANDLE, "invalid resource handle" },
	{ DLM_DEADLOCK, "deadlock recovery refused this request" },
	{ DLM_DENIED_NOASTS, "failed to allocate AST" },
	{ DLM_FORWARD, "request must wait for primary's response" },
	{ DLM_TIMEOUT, "timeout value for lock has expired" },
	{ DLM_IVGROUPID, "invalid group specification" },
	{ DLM_VERS_CONFLICT, "version conflicts prevent request handling" },
	{ DLM_BAD_DEVICE_PATH, "Locks device does not exist or path wrong" },
	{ DLM_NO_DEVICE_PERMISSION, "Client has insufficient perms for device" },
	{ DLM_NO_CONTROL_DEVICE, "Cannot set options on opened device " },
	{ DLM_RECOVERING, "lock resource being recovered" },
	{ DLM_MIGRATING, "lock resource being migrated" },
	{ DLM_MAXSTATS, "invalid error number" },
	{ 0x0000,  NULL }
};

#define DLM_ASSERT_MASTER_MLE_CLEANUP      0x00000001
#define DLM_ASSERT_MASTER_REQUERY          0x00000002
#define DLM_ASSERT_MASTER_FINISH_MIGRATION 0x00000004
static const value_string dlm_assert_master_flags[] = {
	{ DLM_ASSERT_MASTER_MLE_CLEANUP, "cleanup" },
	{ DLM_ASSERT_MASTER_REQUERY, "requery" },
	{ DLM_ASSERT_MASTER_FINISH_MIGRATION, "finish" },
	{ 0x0000,  NULL }
};

#define DLM_FINALIZE_STAGE2  0x01
static const value_string dlm_finalize_reco_flags[] = {
	{ DLM_FINALIZE_STAGE2, "stage2" },
	{ 0x0000,  NULL }
};

enum dlm_ast_type {
	DLM_AST = 0,
	DLM_BAST,
	DLM_ASTUNLOCK
};
static const value_string dlm_proxy_ast_types[] = {
	{ DLM_AST, "AST" },
	{ DLM_BAST, "BAST" },
	{ DLM_ASTUNLOCK, "Unlock AST (unused)" },
	{ 0x0000,  NULL }
};



static void decode_lock_name(proto_tree *tree, guint offset, tvbuff_t *tvb, int len)
{
	guint8 *lck, *lockname;
	enum ocfs2_lock_type t;
	char lt = '?';
	long unsigned int blkno = 0;
	unsigned int generation = 0;
	proto_item *item;
	
	lck = tvb_get_string(tvb, offset, len);
	lockname = g_strndup(lck, len);

#define OCFS2_LOCK_ID_PAD "000000"
#define OCFS2_LOCK_ID_MAX_LEN  32
#define OCFS2_LOCK_ID_FORMAT   "%c"OCFS2_LOCK_ID_PAD"%016"MLFx64"%08x"
	if (lockname &&
	    sscanf(lockname, OCFS2_LOCK_ID_FORMAT, &lt, &blkno, &generation) == 3) {
		t = ocfs2_lock_type_from_lockname(&lt);
		item = proto_tree_add_item(tree, hf_dlm_ocfs2_lock_type, tvb, offset, 1, FALSE);
		proto_item_append_text(item, ": %s", ocfs2_lock_type_to_str(t));
		offset += 1;

		offset += strlen(OCFS2_LOCK_ID_PAD);

		item = proto_tree_add_item(tree, hf_dlm_ocfs2_blkno, tvb, offset, 16, FALSE);
		proto_item_append_text(item, ": %lu", blkno);
		offset += 16;
		
		item = proto_tree_add_item(tree, hf_dlm_ocfs2_generation, tvb, offset, 8, FALSE);
		proto_item_append_text(item, ": %08x", generation);
		//proto_tree_add_text(tree, tvb, offset, len, "lockname='%.*s', type=%s, blkno=%lu, gen=%08X",
		//		    len, lockname, ocfs2_lock_type_to_str(t), blkno, generation);
	}
	g_free(lockname);
}

static void decode_lvb(proto_tree *tree, guint offset, tvbuff_t *tvb)
{
	guint32 oldseq;
	guint32 version;
	guint32 clusters;
	guint32 uid;
	guint32 gid;
	guint16 mode;
	guint16 nlink;
	guint64 raw_atime;
	guint64 raw_ctime;
	guint64 raw_mtime;
	guint64 size;
	guint32 reserved1;
	guint32 reserved2;
	struct passwd *pw;
	struct group *gr;
	guint off = offset;
	struct tm t_atime, t_ctime, t_mtime;
	char atime_buf[32];
	char ctime_buf[32];
	char mtime_buf[32];
	proto_item *item;
	const char *modestr = "unknown";

	oldseq = tvb_get_ntohl(tvb, off); off += 4;

	version = tvb_get_ntohl(tvb, off); off += 4;

	clusters = tvb_get_ntohl(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_clusters, tvb, off, 4, FALSE);
	off += 4;

	uid = tvb_get_ntohl(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_uid, tvb, off, 4, FALSE);
	pw = getpwuid(uid);
	if (pw)
		proto_item_append_text(item, " (%s)", pw->pw_name);
	off += 4;

	gid = tvb_get_ntohl(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_gid, tvb, off, 4, FALSE);
	gr = getgrgid(gid);
	if (gr)
		proto_item_append_text(item, " (%s)", gr->gr_name);
	off += 4;

	mode = tvb_get_ntohs(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_mode, tvb, off, 2, FALSE);
	/* TODO: full mode parsing, drwxr-xr-x style */
	switch (mode & S_IFMT) {
		case S_IFDIR:
			modestr = "directory";
			break;
		case S_IFCHR:
			modestr = "char device";
			break;
		case S_IFBLK:
			modestr = "block device";
			break;
		case S_IFREG:
			modestr = "regular file";
			break;
		case S_IFIFO:
			modestr = "fifo";
			break;
		case S_IFLNK:
			modestr = "symlink";
			break;
		default:
			modestr = "unknown";
			break;
	}
	proto_item_append_text(item, " (%s)", modestr);
	off += 2;

	nlink = tvb_get_ntohs(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_nlink, tvb, off, 2, FALSE);
	off += 2;

	raw_atime = tvb_get_ntoh64(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_atime, tvb, off, 8, FALSE);
	ocfs2_lvb_time_get_localtime(raw_atime, &t_atime);
	strftime(atime_buf, 32, "%Y-%m-%d %H:%M:%S", &t_atime);
	proto_item_append_text(item, " (%s.%09u)", atime_buf, ocfs2_lvb_time_get_nsec(raw_atime));
	off += 8;

	raw_ctime = tvb_get_ntoh64(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_ctime, tvb, off, 8, FALSE);
	ocfs2_lvb_time_get_localtime(raw_ctime, &t_ctime);
	strftime(ctime_buf, 32, "%Y-%m-%d %H:%M:%S", &t_ctime);
	proto_item_append_text(item, " (%s.%09u)", ctime_buf, ocfs2_lvb_time_get_nsec(raw_ctime));
	off += 8;

	raw_mtime = tvb_get_ntoh64(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_mtime, tvb, off, 8, FALSE);
	ocfs2_lvb_time_get_localtime(raw_mtime, &t_mtime);
	strftime(mtime_buf, 32, "%Y-%m-%d %H:%M:%S", &t_mtime);
	proto_item_append_text(item, " (%s.%09u)", mtime_buf, ocfs2_lvb_time_get_nsec(raw_mtime));
	off += 8;

	size = tvb_get_ntoh64(tvb, off);
	item = proto_tree_add_item(tree, hf_dlm_ocfs2_size, tvb, off, 8, FALSE);
	off += 8;

	reserved1 = tvb_get_ntohl(tvb, off); off += 4;

	reserved2 = tvb_get_ntohl(tvb, off); off += 4;



	proto_tree_add_text(tree, tvb, offset, DLM_LVB_LEN, 
			    "oldseq=%u, version=%u, reserved1=%u, reserved2=%u",
			    oldseq, version, reserved1, reserved2);
}




struct dlm_msg_field_def;

struct dlm_msg_field_def
{
	const char *name;
	int *hf_index;
	int offset;
	int len;
	void (*field_handler) (proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld);
};

struct dlm_msg_struct_def
{
	const char *name;
	const char *description;
	guint16 msg_type;
	gint *ett_index;
	struct dlm_msg_field_def fields[12];
};

/* for the rare case of one field needing info from another
 * e.g. name needs to know the value of namelen
 * so namelen should always precede name in the struct (which it does) */
struct dlm_msg_struct_priv
{
	int valid;
	union {
		guint32 namelen;
	} u;
};

#define DLM_MSG_STRUCT_DEF_END       NULL, NULL, 0, 0, NULL


void dlm_am_flags_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint32 flags = tvb_get_ntohl(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_ast_type_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint8 type = tvb_get_guint8(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_cookie_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	/* TODO: break down cookie into node_idx:lock_number */
	//guint64 flags = tvb_get_ntoh64(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_domain_name_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	/* TODO: attempt to match this up with a mounted filesystem somehow */
	//guint8 namelen = tvb_get_guint8(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_domain_namelen_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint8 namelen = tvb_get_guint8(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_fr_flags_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint8 flags = tvb_get_guint8(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_lkm_flags_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	proto_item *item;
	guint32 flags = tvb_get_ntohl(tvb, offset + fld->offset);
	item = proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
	proto_item_append_text(item, ": %s", 
			    decode_enumerated_bitfield_full(flags, 0xffffffff, 32, dlm_lock_flags, "%s"));

}
void dlm_lockmode_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint8 lockmode = tvb_get_guint8(tvb, offset + fld->offset);
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}
void dlm_lvb_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	/* TODO: instead of just text labels, break out each ocfs2 specific lvb field into a separate, filterable field */
	if (tvb_offset_exists(tvb, offset + fld->offset)) {
		proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
		decode_lvb(tree, offset + fld->offset, tvb);
	}
}
void dlm_name_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	guint32 namelen;
	struct dlm_msg_struct_priv *p = (struct dlm_msg_struct_priv *)priv;
	if (p->valid) {
		namelen = p->u.namelen;
		proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, namelen, FALSE);
		decode_lock_name(tree, offset + fld->offset, tvb, namelen);
	}
}
void dlm_namelen_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	guint8 namelen = tvb_get_guint8(tvb, offset + fld->offset);
	struct dlm_msg_struct_priv *p = (struct dlm_msg_struct_priv *)priv;
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
	p->valid = 1;
	p->u.namelen = namelen;

}
void dlm_node_idx_handler(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld)
{
	//guint8 node_idx = tvb_get_guint8(tvb, offset + fld->offset);
	// TODO: look into printing the hostname by working with o2cb stuff
	proto_tree_add_item(tree, *fld->hf_index, tvb, offset + fld->offset, fld->len, FALSE);
}


/* DLM structures */
struct dlm_master_request
{
        guint8 node_idx;
        guint8 namelen;
        guint16 pad1;	// unused
        guint32 flags;  // unused
        guint8 name[O2NM_MAX_NAME_LEN];
};
struct dlm_assert_master
{
	guint8 node_idx;
	guint8 namelen;
	guint16 pad1;	// unused
	guint32 flags;
	guint8 name[O2NM_MAX_NAME_LEN];
};
struct dlm_convert_lock
{
	guint64 cookie;
	guint32 flags;
	guint8 pad1;	// unused
	guint8 node_idx;
	gint8 requested_type;
	guint8 namelen;
	guint8 name[O2NM_MAX_NAME_LEN];
	gint8 lvb[0];  // DLM_LVB_LEN
};
struct dlm_create_lock
{
	guint64 cookie;
	guint32 flags;
	guint8 pad1;	// unused
	guint8 node_idx;
	gint8 requested_type;
	guint8 namelen;
	guint8 name[O2NM_MAX_NAME_LEN];
};
struct dlm_unlock_lock
{
	guint64 cookie;
	guint32 flags;
	guint16 pad1;	// unused
	guint8 node_idx;
	guint8 namelen;
	guint8 name[O2NM_MAX_NAME_LEN];
	gint8 lvb[0];  // DLM_LVB_LEN
};
struct dlm_proxy_ast
{
	guint64 cookie;

	guint32 flags;
	guint8 node_idx;
	guint8 type;
	guint8 blocked_type;
	guint8 namelen;

	guint8 name[O2NM_MAX_NAME_LEN];

	gint8 lvb[0];  // DLM_LVB_LEN
};
struct dlm_lock_request
{
	guint8 node_idx;
	guint8 dead_node;
	guint16 pad1;	// unused
	guint32 pad2;	// unused
};
struct dlm_reco_data_done
{
	guint8 node_idx;
	guint8 dead_node;
	guint16 pad1;	// unused
	guint32 pad2;	// unused
	guint8 reco_lvb[DLM_LVB_LEN]; // unused
};
struct dlm_begin_reco
{
	guint8 node_idx;
	guint8 dead_node;
	guint16 pad1;	// unused
	guint32 pad2;	// unused
};
struct dlm_query_join_request
{
	guint8 node_idx;
	guint8 pad1[2];	// unused
	guint8 name_len;
	guint8 domain[O2NM_MAX_NAME_LEN];
};
struct dlm_assert_joined
{
	guint8 node_idx;
	guint8 pad1[2];	// unused
	guint8 name_len;
	guint8 domain[O2NM_MAX_NAME_LEN];
};
struct dlm_cancel_join
{
	guint8 node_idx;
	guint8 pad1[2];	// unused
	guint8 name_len;
	guint8 domain[O2NM_MAX_NAME_LEN];
};
struct dlm_exit_domain
{
	guint8 node_idx;
	guint8 pad1[3];	// unused
};
struct dlm_finalize_reco
{
	guint8 node_idx;
	guint8 dead_node;
	guint8 flags;
	guint8 pad1;	// unused
	guint32 pad2;	// unused
};

static struct dlm_msg_struct_def dlm_struct_defs[] = {
{ "dlm_master_request", "DLM Master Request", DLM_MASTER_REQUEST_MSG, &ett_dlm_master_request, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_master_request,node_idx),	dlm_node_idx_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_master_request,namelen),	dlm_namelen_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_master_request,name),		dlm_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_assert_master", "DLM Assert Master", DLM_ASSERT_MASTER_MSG, &ett_dlm_assert_master, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_assert_master,node_idx),	dlm_node_idx_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_assert_master,namelen),	dlm_namelen_handler},
	{ "flags",	&hf_dlm_am_flags,	FIELD_OFFSET_AND_SIZE(struct dlm_assert_master,flags),		dlm_am_flags_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_assert_master,name),		dlm_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_create_lock", "DLM Create Lock", DLM_CREATE_LOCK_MSG, &ett_dlm_create_lock, {
	{ "cookie",	&hf_dlm_cookie,		FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,cookie),		dlm_cookie_handler},
	{ "flags",	&hf_dlm_flags,		FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,flags),		dlm_lkm_flags_handler},
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,node_idx),		dlm_node_idx_handler},
	{ "requested_type",&hf_dlm_requested_type, FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,requested_type),dlm_lockmode_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,namelen),		dlm_namelen_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_create_lock,name),		dlm_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_convert_lock", "DLM Convert Lock", DLM_CONVERT_LOCK_MSG, &ett_dlm_convert_lock,  {
	{ "cookie",	&hf_dlm_cookie,		FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,cookie),		dlm_cookie_handler},
	{ "flags",	&hf_dlm_flags,		FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,flags),		dlm_lkm_flags_handler},
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,node_idx),	dlm_node_idx_handler},
	{ "requested_type",&hf_dlm_requested_type, FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,requested_type),dlm_lockmode_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,namelen),		dlm_namelen_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_convert_lock,name),		dlm_name_handler},
	{ "lvb",	&hf_dlm_lvb,		offsetof(struct dlm_convert_lock,lvb[0]), DLM_LVB_LEN,		dlm_lvb_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_unlock_lock", "DLM Unlock Lock", DLM_UNLOCK_LOCK_MSG, &ett_dlm_unlock_lock, {
	{ "cookie",	&hf_dlm_cookie,		FIELD_OFFSET_AND_SIZE(struct dlm_unlock_lock,cookie),		dlm_cookie_handler},
	{ "flags",	&hf_dlm_flags,		FIELD_OFFSET_AND_SIZE(struct dlm_unlock_lock,flags),		dlm_lkm_flags_handler},
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_unlock_lock,node_idx),		dlm_node_idx_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_unlock_lock,namelen),		dlm_namelen_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_unlock_lock,name),		dlm_name_handler},
	{ "lvb",	&hf_dlm_lvb,		offsetof(struct dlm_unlock_lock,lvb[0]), DLM_LVB_LEN,		dlm_lvb_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_proxy_ast", "DLM Proxy AST", DLM_PROXY_AST_MSG, &ett_dlm_proxy_ast, {
	{ "cookie",	&hf_dlm_cookie,		FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,cookie),		dlm_cookie_handler},
	{ "flags",	&hf_dlm_flags,		FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,flags),		dlm_lkm_flags_handler},
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,node_idx),		dlm_node_idx_handler},
	{ "type",	&hf_dlm_ast_type, 	FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,type),		dlm_ast_type_handler},
	{ "blocked_type",&hf_dlm_blocked_type,	FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,blocked_type),	dlm_lockmode_handler},
	{ "namelen",	&hf_dlm_namelen,	FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,namelen),		dlm_namelen_handler},
	{ "name",	&hf_dlm_name,		FIELD_OFFSET_AND_SIZE(struct dlm_proxy_ast,name),		dlm_name_handler},
	{ "lvb",	&hf_dlm_lvb,		offsetof(struct dlm_proxy_ast,lvb[0]), DLM_LVB_LEN,		dlm_lvb_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_lock_request", "DLM Lock Request (recovery)", DLM_LOCK_REQUEST_MSG, &ett_dlm_lock_request, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_lock_request,node_idx),	dlm_node_idx_handler},
	{ "dead_node",	&hf_dlm_dead_node,	FIELD_OFFSET_AND_SIZE(struct dlm_lock_request,dead_node),	dlm_node_idx_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_reco_data_done", "DLM Recovery Data Done (recovery)", DLM_RECO_DATA_DONE_MSG, &ett_dlm_recovery_data_done, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_reco_data_done,node_idx),	dlm_node_idx_handler},
	{ "dead_node",	&hf_dlm_dead_node,	FIELD_OFFSET_AND_SIZE(struct dlm_reco_data_done,dead_node),	dlm_node_idx_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_begin_reco", "DLM Begin Recovery (recovery)", DLM_BEGIN_RECO_MSG, &ett_dlm_begin_recovery, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_begin_reco,node_idx),	dlm_node_idx_handler},
	{ "dead_node",	&hf_dlm_dead_node,	FIELD_OFFSET_AND_SIZE(struct dlm_begin_reco,dead_node),	dlm_node_idx_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_query_join_request", "DLM Query Join Request", DLM_QUERY_JOIN_MSG, &ett_dlm_query_join, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_query_join_request,node_idx),	dlm_node_idx_handler},
	{ "name_len",	&hf_dlm_domain_name_len,FIELD_OFFSET_AND_SIZE(struct dlm_query_join_request,name_len),	dlm_domain_namelen_handler},
	{ "domain",	&hf_dlm_domain_name,	FIELD_OFFSET_AND_SIZE(struct dlm_query_join_request,domain),	dlm_domain_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_assert_joined", "DLM Assert Joined", DLM_ASSERT_JOINED_MSG, &ett_dlm_assert_joined, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_assert_joined,node_idx),	dlm_node_idx_handler},
	{ "name_len",	&hf_dlm_domain_name_len,FIELD_OFFSET_AND_SIZE(struct dlm_assert_joined,name_len),	dlm_domain_namelen_handler},
	{ "domain",	&hf_dlm_domain_name,	FIELD_OFFSET_AND_SIZE(struct dlm_assert_joined,domain),		dlm_domain_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_cancel_join", "DLM Cancel Join", DLM_CANCEL_JOIN_MSG, &ett_dlm_cancel_join, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_cancel_join,node_idx),	dlm_node_idx_handler},
	{ "name_len",	&hf_dlm_domain_name_len,FIELD_OFFSET_AND_SIZE(struct dlm_cancel_join,name_len),	dlm_domain_namelen_handler},
	{ "domain",	&hf_dlm_domain_name,	FIELD_OFFSET_AND_SIZE(struct dlm_cancel_join,domain),	dlm_domain_name_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_exit_domain", "DLM Exit Domain", DLM_EXIT_DOMAIN_MSG, &ett_dlm_exit_domain, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_exit_domain,node_idx),	dlm_node_idx_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
},
{ "dlm_finalize_reco", "DLM Finalize Recovery", DLM_FINALIZE_RECO_MSG, &ett_dlm_finalize_recovery, {
	{ "node_idx",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,node_idx),	dlm_node_idx_handler},
	{ "dead_node",	&hf_dlm_dead_node,	FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,dead_node),	dlm_node_idx_handler},
	{ "flags",	&hf_dlm_fr_flags,	FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,flags),		dlm_fr_flags_handler},
	{ DLM_MSG_STRUCT_DEF_END } }
}
};
	



#if 0

/*
 * unimplemented cuz i forgot about these
 */

struct dlm_migrate_request
{
	guint8 master;
	guint8 new_master;
	guint8 namelen;
	guint8 pad1;
	guint32 pad2;
	guint8 name[O2NM_MAX_NAME_LEN];
};
{ "dlm_migrate_request", "DLM Migrate Request", DLM_MIGRATE_REQUEST_MSG,
	{ "master",	&hf_dlm_node_idx,	FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,node_idx),	dlm_node_idx_handler,	NULL },
	{ "dead_node",	&hf_dlm_dead_node,	FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,dead_node),	dlm_node_idx_handler,	NULL },
	{ "flags",	&hf_dlm_flags,		FIELD_OFFSET_AND_SIZE(struct dlm_finalize_reco,flags),		dlm_fr_flags_handler,	NULL },
	{ DLM_MSG_STRUCT_DEF_END }
},

struct dlm_master_requery
{
	guint8 pad1;
	guint8 pad2;
	guint8 node_idx;
	guint8 namelen;
	guint32 pad3;
	guint8 name[O2NM_MAX_NAME_LEN];
};

struct dlm_migratable_lock
{
	guint64 cookie;

	/* these 3 are just padding for the in-memory structure, but
	 * list and flags are actually used when sent over the wire */
	guint16 pad1;
	guint8 list;  // 0=granted, 1=converting, 2=blocked
	guint8 flags;

	gint8 type;
	gint8 convert_type;
	gint8 highest_blocked;
	guint8 node;
};  // 16 bytes

#define DLM_MRES_RECOVERY   0x01
#define DLM_MRES_MIGRATION  0x02
#define DLM_MRES_ALL_DONE   0x04

/*
 * We would like to get one whole lockres into a single network
 * message whenever possible.  Generally speaking, there will be
 * at most one dlm_lock on a lockres for each node in the cluster,
 * plus (infrequently) any additional locks coming in from userdlm.
 *
 * struct _dlm_lockres_page
 * {
 * 	dlm_migratable_lockres mres;
 * 	dlm_migratable_lock ml[DLM_MAX_MIGRATABLE_LOCKS];
 * 	guint8 pad[DLM_MIG_LOCKRES_RESERVED];
 * };
 *
 * from ../cluster/tcp.h
 *    NET_MAX_PAYLOAD_BYTES  (4096 - sizeof(net_msg))
 *    (roughly 4080 bytes)
 * and sizeof(dlm_migratable_lockres) = 112 bytes
 * and sizeof(dlm_migratable_lock) = 16 bytes
 *
 * Choosing DLM_MAX_MIGRATABLE_LOCKS=240 and
 * DLM_MIG_LOCKRES_RESERVED=128 means we have this:
 *
 *  (DLM_MAX_MIGRATABLE_LOCKS * sizeof(dlm_migratable_lock)) +
 *     sizeof(dlm_migratable_lockres) + DLM_MIG_LOCKRES_RESERVED =
 *        NET_MAX_PAYLOAD_BYTES
 *  (240 * 16) + 112 + 128 = 4080
 *
 * So a lockres would need more than 240 locks before it would
 * use more than one network packet to recover.  Not too bad.
 */
#define DLM_MAX_MIGRATABLE_LOCKS   240

struct dlm_migratable_lockres
{
	guint8 master;
	guint8 lockname_len;
	guint8 num_locks;    // locks sent in this structure
	guint8 flags;
	guint32 total_locks; // locks to be sent for this migration cookie
	guint64 mig_cookie;  // cookie for this lockres migration
			 // or zero if not needed
	// 16 bytes
	guint8 lockname[DLM_LOCKID_NAME_MAX];
	// 48 bytes
	guint8 lvb[DLM_LVB_LEN];
	// 112 bytes
	struct dlm_migratable_lock ml[0];  // 16 bytes each, begins at byte 112
};
#define DLM_MIG_LOCKRES_MAX_LEN  \
	(sizeof(struct dlm_migratable_lockres) + \
	 (sizeof(struct dlm_migratable_lock) * \
	  DLM_MAX_MIGRATABLE_LOCKS) )

/* from above, 128 bytes
 * for some undetermined future use */
#define DLM_MIG_LOCKRES_RESERVED   (NET_MAX_PAYLOAD_BYTES - \
				    DLM_MIG_LOCKRES_MAX_LEN)


#endif


#if 0
/* EEEEK still need to do these 3 */
static gint ett_dlm_migrate_request = -1;
static gint ett_dlm_migrate_lockres = -1;
static gint ett_dlm_master_requery = -1;
#endif


static int valid_magic(guint16 magic)
{
	return magic == O2NET_MSG_MAGIC || 
		magic == O2NET_MSG_STATUS_MAGIC ||
		magic == O2NET_MSG_KEEP_REQ_MAGIC ||
		magic == O2NET_MSG_KEEP_RESP_MAGIC;
}

/* returns -ve bytes needed, or +ve/0 as the offset of the header */
static int ocfs2_find_header(tvbuff_t *tvb, packet_info *pinfo)
{
	guint offset = 0, available;
	guint16 len, magic = 0; /* invalid */
	int ret;

	available = tvb_reported_length_remaining(tvb, offset);

	while (offset + sizeof(magic) <= available) {
		magic = tvb_get_ntohs(tvb, offset);
		if (valid_magic(magic))
			break;
		offset++;
	}

	if (!valid_magic(magic)) {
		ret = -1;
		goto out;
	}

	len = O2NET_MSG_HDR_LEN;
	if (magic == O2NET_MSG_MAGIC && available >= len)
		len += tvb_get_ntohs(tvb, offset + O2NET_MSG_HDR_OFF_DATA_LEN);

	if (len < available)
		ret = available - len;
	else
		ret = len;
out:
	if (ret < 0) {
		pinfo->desegment_offset = offset;
		pinfo->desegment_len = -ret;
	}
	return ret;
}

static struct dlm_msg_struct_def * msg_struct_lookup_by_type(guint16 msg_type)
{
	unsigned int i;
	struct dlm_msg_struct_def *def;

	for (i=0; i<sizeof(dlm_struct_defs)/sizeof(struct dlm_msg_struct_def); i++) {
		def = &dlm_struct_defs[i];
		if (def->msg_type == msg_type)
			return def;
	}
	return NULL;
}

static void msg_struct_enumerate_fields(struct dlm_msg_struct_def *def, proto_tree *tree, tvbuff_t *tvb, guint offset)
{
	struct dlm_msg_field_def *fld;
	struct dlm_msg_struct_priv priv;
	void (*fn)(proto_tree *tree, tvbuff_t *tvb, guint offset, void *priv, struct dlm_msg_field_def *fld); 
	int i = 0;

	memset(&priv, 0, sizeof(priv));
	
	while (1) {
		fld = &def->fields[i];
		/* done? */
		if (fld->name == NULL)
			break;
		fn = fld->field_handler;
		fn(tree, tvb, offset, &priv, fld);
		i++;
	}
}

static int dissect_ocfs2(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	proto_tree *subtree = NULL;
	proto_item *ti;
	guint16 len, magic, msg_type;
	int ret;

	ret = ocfs2_find_header(tvb, pinfo);
	if (ret < 0)
		goto out;

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "OCFS2");

	magic = tvb_get_ntohs(tvb, O2NET_MSG_HDR_OFF_MAGIC);
	msg_type = tvb_get_ntohs(tvb, O2NET_MSG_HDR_OFF_MSG_TYPE);

	if (check_col(pinfo->cinfo, COL_INFO)) {
		col_clear(pinfo->cinfo, COL_INFO);
		switch (magic) {
		case O2NET_MSG_MAGIC:
		case O2NET_MSG_STATUS_MAGIC:
			col_set_str(pinfo->cinfo, COL_INFO,
				    magic == O2NET_MSG_MAGIC? "Q: ": "A: ");
			col_append_str(pinfo->cinfo, COL_INFO,
				       val_to_str(msg_type, dlm_magic, "???"));
			break;
		case O2NET_MSG_KEEP_REQ_MAGIC:
			col_set_str(pinfo->cinfo, COL_INFO, "Q: ping");
			break;
		case O2NET_MSG_KEEP_RESP_MAGIC:
			col_set_str(pinfo->cinfo, COL_INFO, "A: ping");
			break;
		}			
	}

	if (tree) {
		ti = proto_tree_add_protocol_format(tree, proto_ocfs2, tvb, 0,
						    -1,
						    "ocfs2");
		subtree = proto_item_add_subtree(ti, ett_ocfs2);
	}

	if (tree) {
		proto_tree_add_item(subtree, hf_msg_magic,
				    tvb, O2NET_MSG_HDR_OFF_MAGIC, 2,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_data_len,
				    tvb, O2NET_MSG_HDR_OFF_DATA_LEN, 2,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_msg_type,
				    tvb, O2NET_MSG_HDR_OFF_MSG_TYPE, 2,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_pad1,
				    tvb, O2NET_MSG_HDR_OFF_PAD1, 2,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_sys_status,
				    tvb, O2NET_MSG_HDR_OFF_SYS_STATUS, 4,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_status,
				    tvb, O2NET_MSG_HDR_OFF_STATUS, 4,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_key,
				    tvb, O2NET_MSG_HDR_OFF_KEY, 4,
				    FALSE);
		proto_tree_add_item(subtree, hf_msg_msg_num,
				    tvb, O2NET_MSG_HDR_OFF_MSG_NUM, 4,
				    FALSE);

		len = tvb_get_ntohs(tvb, O2NET_MSG_HDR_OFF_DATA_LEN);
		if (len)
			proto_tree_add_item(subtree, hf_msg_payload,
					    tvb, O2NET_MSG_HDR_OFF_PAYLOAD, len,
					    FALSE);

		if (magic == O2NET_MSG_MAGIC) {
			struct dlm_msg_struct_def *def;
			def = msg_struct_lookup_by_type(msg_type);
			if (def)
				msg_struct_enumerate_fields(def, subtree, tvb, O2NET_MSG_HDR_OFF_PAYLOAD);
		}
	}
out:
	return ret;
}

void proto_register_ocfs2(void)
{
	static hf_register_info hf[] = {
		/* ocfs2_msg */
		{ &hf_msg_magic, { "Magic", "ocfs2.msg.magic",
					FT_UINT16, BASE_HEX, VALS(o2net_magic), 0x0,
					"Magic number identifier of O2NET-over-TCPmessage", HFILL } },
		{ &hf_msg_data_len, { "Len", "ocfs2.msg.data_len",
					FT_UINT16, BASE_DEC, NULL, 0x0,
					"Data length", HFILL } },
		{ &hf_msg_msg_type, { "Type","ocfs2.msg.msg_type",
					FT_UINT16, BASE_DEC, VALS(dlm_magic), 0x0,
					"Message type", HFILL } },
		{ &hf_msg_pad1, { "Pad", "ocfs2.msg.pad1",
					FT_UINT16, BASE_HEX, NULL, 0x0,
					"Padding", HFILL } },
		{ &hf_msg_sys_status, { "Sys Status",	"ocfs2.msg.sys_status",
					FT_UINT32, BASE_DEC, VALS(dlm_errnames), 0x0,
					"System level status return code", HFILL } },
		{ &hf_msg_status, { "Status", "ocfs2.msg.status",
					FT_UINT32, BASE_DEC, NULL, 0x0,
					"Return code", HFILL } },
		{ &hf_msg_key, { "Key", "ocfs2.msg.key",
					FT_UINT32, BASE_HEX, NULL, 0x0,
					"Key", HFILL } },
		{ &hf_msg_msg_num, { "Num", "ocfs2.msg.msg_num",
					FT_UINT32, BASE_DEC, NULL, 0x0,
					"Message identification number", HFILL } },
		{ &hf_msg_payload, { "Payload", "ocfs2.msg.payload",
					FT_BYTES, BASE_NONE, NULL, 0x0,
					"Payload", HFILL } },
		{ &hf_dlm_node_idx, { "Node Index", "ocfs2.dlm.node_idx",
					FT_UINT8, BASE_DEC, NULL, 0x0,
					"Node index", HFILL } },
		{ &hf_dlm_flags, { "Flags", "ocfs2.dlm.flags",
					FT_UINT32, BASE_HEX, NULL, 0x0,
					"Flags", HFILL } },
		{ &hf_dlm_am_flags, { "Assert Master Flags", "ocfs2.dlm.am_flags",
					FT_UINT32, BASE_HEX, VALS(dlm_assert_master_flags), 0x0,
					"Assert Master Flags", HFILL } },
		{ &hf_dlm_fr_flags, { "Finalize Recovery Flags", "ocfs2.dlm.fr_flags",
					FT_UINT32, BASE_HEX, VALS(dlm_finalize_reco_flags), 0x0,
					"Finalize Recovery Flags", HFILL } },
		{ &hf_dlm_namelen, { "Namelen", "ocfs2.dlm.namelen",
					FT_UINT8, BASE_DEC, NULL, 0x0,
					"Name length", HFILL } },
		{ &hf_dlm_name, { "Name", "ocfs2.dlm.name",
					FT_STRING, BASE_NONE, NULL, 0x0,
					"Name", HFILL } },
		{ &hf_dlm_cookie, { "Cookie", "ocfs2.dlm.cookie",
					FT_UINT64, BASE_HEX, NULL, 0x0,
					"Unique ID for a single lock on a resource", HFILL } },
		{ &hf_dlm_requested_type, { "Requested Type", "ocfs2.dlm.requested_type",
					FT_UINT8, BASE_DEC, VALS(dlm_lock_modes), 0x0,
					"Requested lock level", HFILL } },
		{ &hf_dlm_blocked_type, { "Blocked Locktype", "ocfs2.dlm.blocked_type",
					FT_UINT8, BASE_DEC, VALS(dlm_lock_modes), 0x0,
					"Blocked lock type", HFILL } },
		{ &hf_dlm_ast_type, { "AST Type", "ocfs2.dlm.ast_type",
					FT_UINT8, BASE_DEC, VALS(dlm_proxy_ast_types), 0x0,
					"AST Type", HFILL } },
		{ &hf_dlm_dead_node, { "Dead Node", "ocfs2.dlm.dead_node",
					FT_UINT8, BASE_DEC, NULL, 0x0, 
					"Dead node index", HFILL } },
		{ &hf_dlm_lvb, { "LVB", "ocfs2.dlm.lvb",
					FT_UINT8, BASE_HEX, NULL, 0x0, 
					"Lock value block", HFILL } },
		{ &hf_dlm_domain_name_len, { "Domain Namelen", "ocfs2.dlm.domain_namelen",
					FT_UINT8, BASE_DEC, NULL, 0x0,
					"Domain name length", HFILL } },
		{ &hf_dlm_domain_name, { "Domain Name", "ocfs2.dlm.domain_name",
					FT_STRING, BASE_NONE, NULL, 0x0,
					"Domain name", HFILL } },

		{ &hf_dlm_ocfs2_blkno, { "blkno", "ocfs2.dlm.blkno",
				FT_STRING, BASE_NONE, NULL, 0x0,
				"Lockname blkno", HFILL } },
		{ &hf_dlm_ocfs2_lock_type, { "lock_type", "ocfs2.dlm.lock_type",
				FT_STRING, BASE_NONE, NULL, 0x0,
				"Lockname lock type", HFILL } },
		{ &hf_dlm_ocfs2_generation, { "generation", "ocfs2.dlm.generation",
				FT_STRING, BASE_NONE, NULL, 0x0,
				"Lockname generation", HFILL } },

		{ &hf_dlm_ocfs2_clusters, { "clusters", "ocfs2.dlm.clusters",
				FT_UINT32, BASE_DEC, NULL, 0x0,
				"LVB clusters", HFILL } },
		{ &hf_dlm_ocfs2_size, { "size", "ocfs2.dlm.size",
				FT_UINT64, BASE_DEC, NULL, 0x0,
				"LVB size", HFILL } },
		{ &hf_dlm_ocfs2_uid, { "uid", "ocfs2.dlm.uid",
				FT_UINT32, BASE_DEC, NULL, 0x0,
				"LVB uid", HFILL } },
		{ &hf_dlm_ocfs2_gid, { "gid", "ocfs2.dlm.gid",
				FT_UINT32, BASE_DEC, NULL, 0x0,
				"LVB gid", HFILL } },
		{ &hf_dlm_ocfs2_mode, { "mode", "ocfs2.dlm.mode",
				FT_UINT16, BASE_HEX, NULL, 0x0,
				"LVB mode", HFILL } },
		{ &hf_dlm_ocfs2_nlink, { "nlink", "ocfs2.dlm.nlink",
				FT_UINT16, BASE_DEC, NULL, 0x0,
				"LVB nlink", HFILL } },
		{ &hf_dlm_ocfs2_atime, { "atime", "ocfs2.dlm.atime",
				FT_UINT64, BASE_DEC, NULL, 0x0,
				"LVB atime", HFILL } },
		{ &hf_dlm_ocfs2_ctime, { "ctime", "ocfs2.dlm.ctime",
				FT_UINT64, BASE_DEC, NULL, 0x0,
				"LVB ctime", HFILL } },
		{ &hf_dlm_ocfs2_mtime, { "mtime", "ocfs2.dlm.mtime",
				FT_UINT64, BASE_DEC, NULL, 0x0,
				"LVB mtime", HFILL } },
	};
	static gint *ett[] = {
		&ett_ocfs2,
		&ett_dlm_master_request,
		&ett_dlm_assert_master,
		&ett_dlm_create_lock,
		&ett_dlm_convert_lock,
		&ett_dlm_proxy_ast,
		&ett_dlm_unlock_lock,
		&ett_dlm_migrate_request,
		&ett_dlm_migrate_lockres,
		&ett_dlm_query_join,
		&ett_dlm_assert_joined,
		&ett_dlm_cancel_join,
		&ett_dlm_exit_domain,
		&ett_dlm_master_requery,
		&ett_dlm_lock_request,
		&ett_dlm_recovery_data_done,
		&ett_dlm_begin_recovery,
		&ett_dlm_finalize_recovery,
	};

	proto_ocfs2 = proto_register_protocol("OCFS2 Networking", "OCFS2",
					      "ocfs2");
	proto_register_field_array(proto_ocfs2, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}

void proto_reg_handoff_ocfs2(void)
{
	dissector_handle_t ocfs2_handle;

	ocfs2_handle = new_create_dissector_handle(dissect_ocfs2, proto_ocfs2);

	dissector_add_uint("tcp.port", TCP_PORT_OCFS2, ocfs2_handle);
}
